# This file is part of the MapProxy project.
# Copyright (C) 2011 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import codecs
import datetime
import json
import socket
import time
import hashlib
import base64
from io import BytesIO

from mapproxy.image import ImageSource
from mapproxy.cache.base import (
    TileCacheBase,
    tile_buffer, CacheBackendError,)
from mapproxy.source import SourceError
from mapproxy.srs import SRS

from threading import Lock

try:
    import requests
except ImportError:
    requests = None


import logging
log = logging.getLogger(__name__)


class UnexpectedResponse(CacheBackendError):
    pass


class CouchDBCache(TileCacheBase):
    def __init__(self, url, db_name,
                 file_ext, tile_grid, md_template=None,
                 tile_id_template=None, coverage=None):
        super(CouchDBCache, self).__init__(coverage)

        if requests is None:
            raise ImportError("CouchDB backend requires 'requests' package.")

        md5 = hashlib.new('md5', (url + db_name).encode('utf-8'), usedforsecurity=False)
        self.lock_cache_id = 'couchdb-' + md5.hexdigest()
        self.file_ext = file_ext
        self.tile_grid = tile_grid
        self.md_template = md_template
        self.couch_url = '%s/%s' % (url.rstrip('/'), db_name.lower())
        self.req_session = requests.Session()
        self.req_session.timeout = 5
        self.db_initialised = False
        self.app_init_db_lock = Lock()
        self.tile_id_template = tile_id_template

    def init_db(self):
        with self.app_init_db_lock:
            if self.db_initialised:
                return
            try:
                self.req_session.put(self.couch_url)
                self.db_initialised = True
            except requests.exceptions.RequestException as ex:
                log.warning('unable to initialize CouchDB: %s', ex)

    def tile_url(self, coord):
        return self.document_url(coord) + '/tile'

    def document_url(self, coord, relative=False):
        x, y, z = coord
        grid_name = self.tile_grid.name
        couch_url = self.couch_url
        if relative:
            if self.tile_id_template:
                if self.tile_id_template.startswith('%(couch_url)s/'):
                    tile_id_template = self.tile_id_template[len('%(couch_url)s/'):]
                else:
                    tile_id_template = self.tile_id_template
                return tile_id_template % locals()
            else:
                return '%(grid_name)s-%(z)s-%(x)s-%(y)s' % locals()
        else:
            if self.tile_id_template:
                return self.tile_id_template % locals()
            else:
                return '%(couch_url)s/%(grid_name)s-%(z)s-%(x)s-%(y)s' % locals()

    def is_cached(self, tile, dimensions=None):
        if tile.coord is None or tile.source:
            return True
        url = self.document_url(tile.coord)
        try:
            self.init_db()
            resp = self.req_session.get(url)
            if resp.status_code == 200:
                doc = json.loads(codecs.decode(resp.content, 'utf-8'))
                tile.timestamp = doc.get(self.md_template.timestamp_key)
                return True
        except (requests.exceptions.RequestException, socket.error) as ex:
            # is_cached should not fail (would abort seeding for example),
            # so we catch these errors here and just return False
            log.warning('error while requesting %s: %s', url, ex)
            return False
        if resp.status_code == 404:
            return False
        raise SourceError('%r: %r' % (resp.status_code, resp.content))

    def _tile_doc(self, tile):
        tile_id = self.document_url(tile.coord, relative=True)
        if self.md_template:
            tile_doc = self.md_template.doc(tile, self.tile_grid)
        else:
            tile_doc = {}
        tile_doc['_id'] = tile_id

        with tile_buffer(tile) as buf:
            data = buf.read()
        tile_doc['_attachments'] = {
            'tile': {
                'content_type': 'image/' + self.file_ext,
                'data': codecs.decode(
                    base64.b64encode(data).replace(b'\n', b''),
                    'ascii',
                ),
            }
        }
        return tile_id, tile_doc

    def _store_bulk(self, tiles):
        tile_docs = {}
        for tile in tiles:
            tile_id, tile_doc = self._tile_doc(tile)
            tile_docs[tile_id] = tile_doc

        duplicate_tiles = self._post_bulk(tile_docs)

        if duplicate_tiles:
            self._fill_rev_ids(duplicate_tiles)
            self._post_bulk(duplicate_tiles, no_conflicts=True)

        return True

    def _post_bulk(self, tile_docs, no_conflicts=False):
        """
        POST multiple tiles, returns all tile docs with conflicts during POST.
        """
        doc = {'docs': list(tile_docs.values())}
        data = json.dumps(doc)
        self.init_db()
        resp = self.req_session.post(self.couch_url + '/_bulk_docs', data=data,
                                     headers={'Content-type': 'application/json'})
        if resp.status_code != 201:
            raise UnexpectedResponse('got unexpected resp (%d) from CouchDB: %s' % (resp.status_code, resp.content))

        resp_doc = json.loads(codecs.decode(resp.content, 'utf-8'))
        duplicate_tiles = {}
        for tile in resp_doc:
            if tile.get('error', 'false') == 'conflict':
                duplicate_tiles[tile['id']] = tile_docs[tile['id']]

        if no_conflicts and duplicate_tiles:
            raise UnexpectedResponse('got unexpected resp (%d) from CouchDB: %s' % (resp.status_code, resp.content))

        return duplicate_tiles

    def _fill_rev_ids(self, tile_docs):
        """
        Request all revs for tile_docs and insert it into the tile_docs.
        """
        keys_doc = {'keys': list(tile_docs.keys())}
        data = json.dumps(keys_doc)
        self.init_db()
        resp = self.req_session.post(self.couch_url + '/_all_docs', data=data,
                                     headers={'Content-type': 'application/json'})
        if resp.status_code != 200:
            raise UnexpectedResponse('got unexpected resp (%d) from CouchDB: %s' % (resp.status_code, resp.content))

        resp_doc = json.loads(codecs.decode(resp.content, 'utf-8'))
        for tile in resp_doc['rows']:
            tile_docs[tile['id']]['_rev'] = tile['value']['rev']

    def store_tile(self, tile, dimensions=None):
        if tile.stored:
            return True

        return self._store_bulk([tile])

    def store_tiles(self, tiles, dimensions=None):
        tiles = [t for t in tiles if not t.stored]
        return self._store_bulk(tiles)

    def load_tile_metadata(self, tile, dimensions=None):
        if tile.timestamp:
            return

        # is_cached loads metadata
        self.is_cached(tile, dimensions=None)

    def load_tile(self, tile, with_metadata=False, dimensions=None):
        # bulk loading with load_tiles is not implemented, because
        # CouchDB's /all_docs? does not include attachments

        if tile.source or tile.coord is None:
            return True
        url = self.document_url(tile.coord) + '?attachments=true'
        self.init_db()
        resp = self.req_session.get(url, headers={'Accept': 'application/json'})
        if resp.status_code == 200:
            doc = json.loads(codecs.decode(resp.content, 'utf-8'))
            tile_data = BytesIO(base64.b64decode(doc['_attachments']['tile']['data']))
            tile.source = ImageSource(tile_data)
            tile.timestamp = doc.get(self.md_template.timestamp_key)
            return True
        return False

    def remove_tile(self, tile, dimensions=None):
        if tile.coord is None:
            return True
        url = self.document_url(tile.coord)
        resp = requests.head(url)
        if resp.status_code == 404:
            # already removed
            return True
        rev_id = resp.headers['etag']
        url += '?rev=' + rev_id.strip('"')
        self.init_db()
        resp = self.req_session.delete(url)
        if resp.status_code == 200:
            return True
        return False


def utc_now_isoformat():
    now = datetime.datetime.utcnow()
    now = now.isoformat()
    # remove milliseconds, add Zulu timezone
    now = now.rsplit('.', 1)[0] + 'Z'
    return now


class CouchDBMDTemplate(object):
    def __init__(self, attributes):
        self.attributes = attributes
        for key, value in attributes.items():
            if value == '{{timestamp}}':
                self.timestamp_key = key
                break
        else:
            attributes['timestamp'] = '{{timestamp}}'
            self.timestamp_key = 'timestamp'

    def doc(self, tile, grid):
        doc = {}
        x, y, z = tile.coord
        for key, value in self.attributes.items():
            if not isinstance(value, str) or not value.startswith('{{'):
                doc[key] = value
                continue

            if value == '{{timestamp}}':
                doc[key] = time.time()
            elif value == '{{x}}':
                doc[key] = x
            elif value == '{{y}}':
                doc[key] = y
            elif value in ('{{z}}', '{{level}}'):
                doc[key] = z
            elif value == '{{utc_iso}}':
                doc[key] = utc_now_isoformat()
            elif value == '{{wgs_tile_centroid}}':
                tile_bbox = grid.tile_bbox(tile.coord)
                centroid = (
                    tile_bbox[0] + (tile_bbox[2]-tile_bbox[0])/2,
                    tile_bbox[1] + (tile_bbox[3]-tile_bbox[1])/2
                )
                centroid = grid.srs.transform_to(SRS(4326), centroid)
                doc[key] = centroid
            elif value == '{{tile_centroid}}':
                tile_bbox = grid.tile_bbox(tile.coord)
                centroid = (
                    tile_bbox[0] + (tile_bbox[2]-tile_bbox[0])/2,
                    tile_bbox[1] + (tile_bbox[3]-tile_bbox[1])/2
                )
                doc[key] = centroid
            else:
                raise ValueError('unknown CouchDB tile_metadata value: %r' % (value, ))
        return doc



# This file is part of the MapProxy project.
# Copyright (C) 2012 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from mapproxy.client.tile import TileClient, TileURLTemplate
from mapproxy.grid import TileGrid
from mapproxy.srs import SRS
from mapproxy.source.tile import TiledSource
from mapproxy.source.error import HTTPSourceErrorHandler
from mapproxy.layer import MapQuery
from mapproxy.test.http import mock_httpd


TEST_SERVER_ADDRESS = ("127.0.0.1", 56413)
TESTSERVER_URL = ("http://%s:%d" % TEST_SERVER_ADDRESS) + "/%(tms_path)s.png"


class TestTileClientOnError(object):

    def setup_method(self):
        self.grid = TileGrid(SRS(4326), bbox=[-180, -90, 180, 90])
        self.client = TileClient(TileURLTemplate(TESTSERVER_URL))

    def test_cacheable_response(self):
        error_handler = HTTPSourceErrorHandler()
        error_handler.add_handler(500, (255, 0, 0), cacheable=True)
        self.source = TiledSource(self.grid, self.client, error_handler=error_handler)

        with mock_httpd(
            TEST_SERVER_ADDRESS,
            [
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 500,
                        "headers": {"content-type": "text/plain"},
                    },
                )
            ],
        ):
            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 0))]

    def test_image_response(self):
        error_handler = HTTPSourceErrorHandler()
        error_handler.add_handler(500, (255, 0, 0), cacheable=False)
        self.source = TiledSource(self.grid, self.client, error_handler=error_handler)

        with mock_httpd(
            TEST_SERVER_ADDRESS,
            [
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 500,
                        "headers": {"content-type": "text/plain"},
                    },
                )
            ],
        ):
            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert not resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 0))]

    def test_multiple_image_responses(self):
        error_handler = HTTPSourceErrorHandler()
        error_handler.add_handler(500, (255, 0, 0), cacheable=False)
        error_handler.add_handler(204, (255, 0, 127, 200), cacheable=True)
        self.source = TiledSource(self.grid, self.client, error_handler=error_handler)

        with mock_httpd(
            TEST_SERVER_ADDRESS,
            [
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 500,
                        "headers": {"content-type": "text/plain"},
                    },
                ),
                (
                    {"path": "/1/0/0.png"},
                    {
                        "body": b"error",
                        "status": 204,
                        "headers": {"content-type": "text/plain"},
                    },
                ),
            ],
        ):

            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert not resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 0))]

            resp = self.source.get_map(
                MapQuery([-180, -90, 0, 90], (256, 256), SRS(4326), format="png")
            )
            assert resp.cacheable
            assert resp.as_image().getcolors() == [((256 * 256), (255, 0, 127, 200))]



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, division

import re
import sys
import functools

from io import BytesIO

import pytest

from mapproxy.image import ImageSource
from mapproxy.srs import SRS
from mapproxy.compat.image import Image
from mapproxy.request.wms import (
    WMS100MapRequest,
    WMS111MapRequest,
    WMS130MapRequest,
    WMS111FeatureInfoRequest,
    WMS111CapabilitiesRequest,
    WMS130CapabilitiesRequest,
    WMS100CapabilitiesRequest,
    WMS100FeatureInfoRequest,
    WMS130FeatureInfoRequest,
    WMS110MapRequest,
    WMS110FeatureInfoRequest,
    WMS110CapabilitiesRequest,
    wms_request,
)
from mapproxy.test.image import is_jpeg, is_png, tmp_image, create_tmp_image
from mapproxy.test.unit.test_image import assert_geotiff_tags
from mapproxy.test.http import mock_httpd
from mapproxy.test.helper import validate_with_dtd, validate_with_xsd
from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "layer.yaml"


class TestBase(SysTest):

    def test_invalid_url(self, app):
        app.get("/invalid?fop", status=404)


class TestCoverageWMS(SysTest):
    config_file = "layer.yaml"

    def test_unknown_version_110(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=1.1.0"
        )
        assert is_110_capa(resp.lxml)

    def test_unknown_version_113(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=1.1.3"
        )
        assert is_111_capa(resp.lxml)

    def test_unknown_version_090(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&WMTVER=0.9.0"
        )
        assert is_100_capa(resp.lxml)

    def test_unknown_version_200(self, app):
        resp = app.get(
            "http://localhost/service?SERVICE=WMS&REQUEST=GetCapabilities"
            "&VERSION=2.0.0"
        )
        assert is_130_capa(resp.lxml)


def bbox_srs_from_boundingbox(bbox_elem):
    return [
        float(bbox_elem.attrib["minx"]),
        float(bbox_elem.attrib["miny"]),
        float(bbox_elem.attrib["maxx"]),
        float(bbox_elem.attrib["maxy"]),
    ]


class TestWMS111(SysTest):
    config_file = "layer.yaml"

    def setup_method(self):
        # WMSTest.setup_method(self)
        self.common_req = WMS111MapRequest(
            url="/service?", param=dict(service="WMS", version="1.1.1")
        )
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )
        self.common_fi_req = WMS111FeatureInfoRequest(
            url="/service?",
            param=dict(
                x="10",
                y="20",
                width="200",
                height="200",
                layers="wms_cache",
                format="image/png",
                query_layers="wms_cache",
                styles="",
                bbox="1000,400,2000,1400",
                srs="EPSG:900913",
            ),
        )

    def test_invalid_request_type(self, app):
        req = str(self.common_map_req).replace("GetMap", "invalid")
        resp = app.get(req, expect_errors=True)
        is_111_exception(resp.lxml, "unknown WMS request type 'invalid'")

    def test_endpoints(self, app):
        for endpoint in ("service", "ows", "wms"):
            req = WMS111CapabilitiesRequest(
                url="/%s?" % endpoint
            ).copy_with_request_params(self.common_req)
            resp = app.get(req)
            assert resp.content_type == "application/vnd.ogc.wms_xml"
            xml = resp.lxml
            assert validate_with_dtd(xml, dtd_name="wms/1.1.1/WMS_MS_Capabilities.dtd")

    def test_wms_capabilities(self, app):
        req = WMS111CapabilitiesRequest(url="/service?").copy_with_request_params(
            self.common_req
        )
        resp = app.get(req)
        assert resp.content_type == "application/vnd.ogc.wms_xml"
        xml = resp.lxml
        assert (
            xml.xpath(
                "//GetMap//OnlineResource/@xlink:href",
                namespaces=dict(xlink="http://www.w3.org/1999/xlink"),
            )[0] ==
            "http://localhost/service?"
        )

        # test for MetadataURL
        assert (
            xml.xpath(
                "//Layer/MetadataURL/OnlineResource/@xlink:href",
                namespaces=dict(xlink="http://www.w3.org/1999/xlink"),
            )[0] ==
            "http://some.url/"
        )
        assert xml.xpath("//Layer/MetadataURL/@type")[0] == "TC211"

        layer_names = set(xml.xpath("//Layer/Layer/Name/text()"))
        expected_names = set(
            [
                "direct_fwd_params",
                "direct",
                "wms_cache",
                "wms_cache_100",
                "wms_cache_130",
                "wms_cache_transparent",
                "wms_merge",
                "tms_cache",
                "tms_fi_cache",
                "wms_cache_multi",
                "wms_cache_link_single",
                "wms_cache_110",
                "watermark_cache",
                "wms_managed_cookies_cache",
            ]
        )
        assert layer_names == expected_names
        assert set(xml.xpath("//Layer/Layer[3]/Abstract/text()")) == set(["Some abstract"])

        bboxs = xml.xpath("//Layer/Layer[1]/BoundingBox")
        bboxs = dict((e.attrib["SRS"], e) for e in bboxs)
        assert_almost_equal_bbox(
            bbox_srs_from_boundingbox(bboxs["EPSG:3857"]),
            [-20037508.3428, -15538711.0963, 18924313.4349, 15538711.0963],
        )
        assert_almost_equal_bbox(
            bbox_srs_from_boundingbox(bboxs["EPSG:4326"]), [-180.0, -70.0, 170.0, 80.0]
        )

        bbox_srs = xml.xpath("//Layer/Layer/BoundingBox")
        bbox_srs = set(e.attrib["SRS"] for e in bbox_srs)
        # we have a coverage in EPSG:4258, but it is not in wms.srs (#288)
        assert "EPSG:4258" not in bbox_srs

        assert validate_with_dtd(xml, dtd_name="wms/1.1.1/WMS_MS_Capabilities.dtd")

    def test_invalid_layer(self, app):
        self.common_map_req.params["layers"] = "invalid"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        is_111_exception(resp.lxml, "unknown layer: invalid", "LayerNotDefined")

    def test_invalid_layer_img_exception(self, app):
        self.common_map_req.params["layers"] = "invalid"
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_invalid_format(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        is_111_exception(
            resp.lxml, "unsupported image format: image/ascii", "InvalidFormat"
        )

    def test_invalid_format_img_exception(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_invalid_format_options_img_exception(self, app):
        self.common_map_req.params["format"] = "image/png; mode=12bit"
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_missing_format_img_exception(self, app):
        del self.common_map_req.params["format"]
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_invalid_srs(self, app):
        self.common_map_req.params["srs"] = "EPSG:1234"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        is_111_exception(resp.lxml, "unsupported srs: EPSG:1234", "InvalidSRS")

    def test_get_map_unknown_style(self, app):
        self.common_map_req.params["styles"] = "unknown"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        is_111_exception(resp.lxml, "unsupported styles: unknown", "StyleNotDefined")

    def test_get_map_too_large(self, app):
        self.common_map_req.params.size = (5000, 5000)
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req, expect_errors=True)
        # is xml, even if inimage was requested
        assert resp.content_type == "application/vnd.ogc.se_xml"
        is_111_exception(resp.lxml, "image size too large")

    def test_get_map_default_style(self, app, fixture_cache_data):
        self.common_map_req.params["styles"] = "default"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        assert Image.open(data).mode == "RGB"

    def test_get_map_png(self, app, fixture_cache_data):
        resp = app.get(self.common_map_req)
        assert "Cache-Control" not in resp.headers
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        assert Image.open(data).mode == "RGB"

    def test_get_map_float_size(self, app, fixture_cache_data):
        self.common_map_req.params['width'] = '200.0'
        resp = app.get(self.common_map_req)
        assert "Cache-Control" not in resp.headers
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        assert Image.open(data).mode == "RGB"

    def test_get_map_png8_custom_format(self, app, fixture_cache_data):
        self.common_map_req.params["layers"] = "wms_cache"
        self.common_map_req.params["format"] = "image/png; mode=8bit"
        resp = app.get(self.common_map_req)
        assert resp.headers["Content-type"] == "image/png; mode=8bit"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "P"

    def test_get_map_png_transparent_non_transparent_data(
        self, app, fixture_cache_data
    ):
        self.common_map_req.params["transparent"] = "True"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGB"

    def test_get_map_png_transparent(self, app, fixture_cache_data):
        self.common_map_req.params["layers"] = "wms_cache_transparent"
        self.common_map_req.params["transparent"] = "True"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        assert Image.open(data).mode == "RGBA"

    def test_get_map_png_w_default_bgcolor(self, app, fixture_cache_data):
        self.common_map_req.params["layers"] = "wms_cache_transparent"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGB"
        assert img.getcolors()[0][1] == (255, 255, 255)

    def test_get_map_png_w_bgcolor(self, app, fixture_cache_data):
        self.common_map_req.params["layers"] = "wms_cache_transparent"
        self.common_map_req.params["bgcolor"] = "0xff00a0"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        img = Image.open(data)
        assert img.mode == "RGB"
        assert sorted(img.getcolors())[-1][1] == (255, 0, 160)

    def test_get_map_jpeg(self, app, fixture_cache_data):
        self.common_map_req.params["format"] = "image/jpeg"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/jpeg"
        assert is_jpeg(BytesIO(resp.body))

    def test_get_map_geotiff(self, app, fixture_cache_data):
        self.common_map_req.params["format"] = "image/tiff"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/tiff"
        img = ImageSource(BytesIO(resp.body)).as_image()
        assert_geotiff_tags(img, (-180, 80), (180/200.0, 80/200.0), 4326, False)

    def test_get_map_xml_exception(self, app):
        self.common_map_req.params["bbox"] = "0,0,90,90"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/ServiceException/@code") == []
        assert "No response from URL" in xml.xpath("//ServiceException/text()")[0]
        assert validate_with_dtd(xml, "wms/1.1.1/exception_1_1_1.dtd")

    def test_direct_layer_error(self, app):
        self.common_map_req.params["layers"] = "direct"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/ServiceException/@code") == []
        # TODO hide error
        # assert 'unable to get map for layers: direct' in \
        #     xml.xpath('//ServiceException/text()')[0]
        assert "No response from URL" in xml.xpath("//ServiceException/text()")[0]

        assert validate_with_dtd(xml, "wms/1.1.1/exception_1_1_1.dtd")

    def test_direct_layer_non_image_response(self, app):
        self.common_map_req.params["layers"] = "direct"
        expected_req = (
            {
                "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326&styles="
                "&VERSION=1.1.1&BBOX=-180.0,0.0,0.0,80.0"
                "&WIDTH=200"
            },
            {"body": b"notanimage", "headers": {"content-type": "image/jpeg"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            resp = app.get(self.common_map_req, expect_errors=True)
            assert resp.content_type == "application/vnd.ogc.se_xml"
            xml = resp.lxml
            assert xml.xpath("/ServiceExceptionReport/ServiceException/@code") == []
            assert (
                "error while processing image file"
                in xml.xpath("//ServiceException/text()")[0]
            )

            assert validate_with_dtd(xml, "wms/1.1.1/exception_1_1_1.dtd")

    def test_get_map_non_image_response(self, app, cache_dir):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                "&VERSION=1.1.1&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                "&WIDTH=256"
            },
            {"body": b"notanimage", "headers": {"content-type": "image/jpeg"}},
        )
        with mock_httpd(
            ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
        ):
            self.common_map_req.params["bbox"] = "0,0,180,90"
            resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"

        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/ServiceException/@code") == []
        assert (
            "unable to transform image: cannot identify image file"
            in xml.xpath("//ServiceException/text()")[0]
        )

        assert validate_with_dtd(xml, "wms/1.1.1/exception_1_1_1.dtd")

        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()

    @pytest.mark.flaky(reruns=5, reruns_delay=2)
    def test_get_map(self, app, base_dir, cache_dir):
        # check global tile lock directory
        tiles_lock_dir = base_dir.join("wmscachetilelockdir")
        # make sure global tile_lock_dir was ot created by other tests
        if tiles_lock_dir.check():
            tiles_lock_dir.remove()

        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"
                resp = app.get(self.common_map_req)
                assert 35000 < int(resp.headers["Content-length"]) < 75000
                assert resp.content_type == "image/png"

        # check custom tile_lock_dir
        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()
        assert tiles_lock_dir.check()

    def test_get_map_direct_fwd_params_layer(self, app):
        img = create_tmp_image((200, 200), format="png")
        expected_req = (
            {
                "path": r"/service?LAYERs=bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326&styles="
                "&VERSION=1.1.1&BBOX=-180.0,0.0,0.0,80.0"
                "&WIDTH=200&TIME=20041012"
            },
            {"body": img},
        )
        with mock_httpd(
            ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
        ):
            self.common_map_req.params["layers"] = "direct_fwd_params"
            self.common_map_req.params["time"] = "20041012"
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"

    def test_get_map_use_direct_from_level(self, app):
        with tmp_image((200, 200), format="png") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                    "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A4326&styles="
                    "&VERSION=1.1.1&BBOX=5.0,-10.0,6.0,-9.0"
                    "&WIDTH=200"
                },
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "5,-10,6,-9"
                resp = app.get(self.common_map_req)
                img.seek(0)
                assert resp.body == img.read()
                is_png(img)
                assert resp.content_type == "image/png"

    def test_get_map_use_direct_from_level_with_transform(self, app):
        with tmp_image((200, 200), format="png") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                    "&REQUEST=GetMap&HEIGHT=200&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=908822.944624,7004479.85652,920282.144964,7014491.63726"
                    "&WIDTH=229"
                },
                {"body": img.read(), "headers": {"content-type": "image/png"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params[
                    "bbox"
                ] = "444122.311736,5885498.04243,450943.508884,5891425.10484"
                self.common_map_req.params["srs"] = "EPSG:25832"
                resp = app.get(self.common_map_req)
                img.seek(0)
                assert resp.body != img.read()
                is_png(img)
                assert resp.content_type == "image/png"

    def test_get_map_invalid_bbox(self, app):
        # min x larger than max x
        url = (
            """/service?SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&BBOX=7,2,-9,10&SRS=EPSG:4326&WIDTH=164&HEIGHT=388&LAYERS=wms_cache&STYLES=&FORMAT=image/png&TRANSPARENT=TRUE"""  # noqa
        )
        resp = app.get(url, expect_errors=True)
        is_111_exception(resp.lxml, "invalid bbox 7,2,-9,10")

    def test_get_map_invalid_bbox2(self, app):
        # broken bbox for the requested srs
        url = (
            """/service?SERVICE=WMS&VERSION=1.1.1&REQUEST=GetMap&BBOX=-72988843.697212,-255661507.634227,142741550.188860,255661507.634227&SRS=EPSG:25833&WIDTH=164&HEIGHT=388&LAYERS=wms_cache_100&STYLES=&FORMAT=image/png&TRANSPARENT=TRUE"""  # noqa
        )
        resp = app.get(url, expect_errors=True)
        # result depends on proj version
        is_111_exception(
            resp.lxml,
            re_msg="Request too large or invalid BBOX.|Could not transform BBOX: Invalid result.",
        )

    def test_get_map_broken_bbox(self, app):
        url = (
            """/service?SERVICE=WMS&VERSION=1.1.11&REQUEST=GetMap&SRS=EPSG:31468&BBOX=-20000855.0573254,2847125.18913603,-19329367.42767611,4239924.78564583&WIDTH=130&HEIGHT=62&LAYERS=wms_cache&STYLES=&FORMAT=image/png&TRANSPARENT=TRUE"""  # noqa
        )
        resp = app.get(url, expect_errors=True)
        is_111_exception(resp.lxml, "Could not transform BBOX: Invalid result.")

    def test_get_map100(self, app, base_dir, cache_dir):
        # check global tile lock directory
        tiles_lock_dir = base_dir.join("defaulttilelockdir")
        # make sure global tile_lock_dir was ot created by other tests
        if tiles_lock_dir.check():
            tiles_lock_dir.remove()

        # request_format tiff, cache format jpeg, wms request in png
        with tmp_image((256, 256), format="tiff") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&FORMAT=TIFF"
                    "&REQUEST=map&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&WMTVER=1.0.0&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/tiff"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"
                self.common_map_req.params["layers"] = "wms_cache_100"
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/png"

        # check global tile lock directory was created
        assert tiles_lock_dir.check()

        assert cache_dir.join(
            "wms_cache_100_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()

    def test_get_map130(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&CRS=EPSG%3A900913&styles="
                    "&VERSION=1.3.0&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"
                self.common_map_req.params["layers"] = "wms_cache_130"
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/png"
        assert cache_dir.join(
            "wms_cache_130_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()

    def test_get_map130_axis_order(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            img = img.read()
            expected_reqs = [
                (
                    {
                        "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                        "&REQUEST=GetMap&HEIGHT=256&CRS=EPSG%3A4326&styles="
                        "&VERSION=1.3.0&BBOX=0.0,90.0,90.0,180.0"
                        "&WIDTH=256"
                    },
                    {"body": img, "headers": {"content-type": "image/jpeg"}},
                )
            ]
            with mock_httpd(("localhost", 42423), expected_reqs):
                self.common_map_req.params["bbox"] = "90,0,180,90"
                self.common_map_req.params["layers"] = "wms_cache_multi"
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/png"
        assert cache_dir.join(
            "wms_cache_multi_EPSG4326/02/000/000/003/000/000/001.jpeg"
        ).check()

    def test_get_featureinfo(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
                "&VERSION=1.1.1&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=10&Y=20&feature_count=100"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            self.common_fi_req.params["feature_count"] = 100
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_coverage(self, app):
        self.common_fi_req.params["bbox"] = "-180.0,-90.0,180.0,90.0"
        self.common_fi_req.params["srs"] = "EPSG:4326"
        self.common_fi_req.params["width"] = "400"
        self.common_fi_req.params["height"] = "200"
        self.common_fi_req.params["x"] = 395  # outside of coverage
        self.common_fi_req.params["y"] = 50
        self.common_fi_req.params["layers"] = 'tms_fi_cache'
        self.common_fi_req.params["query_layers"] = 'tms_fi_cache'

        resp = app.get(self.common_fi_req)
        assert resp.body == b""
        assert resp.content_type == "text/plain"

        expected_req = (
            {
                "path": r"/service?LAYERs=fi&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A4326"
                "&VERSION=1.1.1&BBOX=-180.0,-90.0,180.0,90.0&styles="
                "&WIDTH=400&QUERY_LAYERS=fi&X=380&Y=50"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            self.common_fi_req.params["x"] = 380  # inside of coverage

            resp = app.get(self.common_fi_req)
            assert resp.body == b"info"
            assert resp.content_type == "text/plain"

    def test_get_featureinfo_float(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
                "&VERSION=1.1.1&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=10.123&Y=20.567&feature_count=100"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            self.common_fi_req.params["feature_count"] = 100
            self.common_fi_req.params["x"] = 10.123
            self.common_fi_req.params["y"] = 20.567
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_transformed(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
                "&BBOX=1172272.30156,7196018.03449,1189711.04571,7213496.99738"
                "&styles=&VERSION=1.1.1&feature_count=100"
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=14&Y=20"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )

        # out fi point at x=10,y=20
        p_25832 = (
            600000 + 10 * (610000 - 600000) / 200,
            6010000 - 20 * (6010000 - 6000000) / 200,
        )
        # the transformed fi point at x=14,y=20
        p_900913 = (
            1172272.30156 + 14 * (1189711.04571 - 1172272.30156) / 200,
            7213496.99738 - 20 * (7213496.99738 - 7196018.03449) / 200,
        )

        # are they the same?
        # check with tolerance: pixel resolution is ~50 and x/y position is rounded to pizel
        assert abs(SRS(25832).transform_to(SRS(900913), p_25832)[0] - p_900913[0]) < 50
        assert abs(SRS(25832).transform_to(SRS(900913), p_25832)[1] - p_900913[1]) < 50

        with mock_httpd(
            ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
        ):
            self.common_fi_req.params["bbox"] = "600000,6000000,610000,6010000"
            self.common_fi_req.params["srs"] = "EPSG:25832"
            self.common_fi_req.params.pos = 10, 20
            self.common_fi_req.params["feature_count"] = 100
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_info_format(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
                "&VERSION=1.1.1&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=10&Y=20"
                "&info_format=text%2Fhtml"
            },
            {"body": b"info", "headers": {"content-type": "text/html"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            self.common_fi_req.params["info_format"] = "text/html"
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/html"
            assert resp.body == b"<html><body><p>info</p></body></html>"

    def test_get_featureinfo_130(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&CRS=EPSG%3A900913"
                "&VERSION=1.3.0&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&I=10&J=20"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            self.common_fi_req.params["layers"] = "wms_cache_130"
            self.common_fi_req.params["query_layers"] = "wms_cache_130"
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_missing_params(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
                "&VERSION=1.1.1&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=10&Y=20"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            del self.common_fi_req.params["format"]
            del self.common_fi_req.params["styles"]
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_missing_params_strict(self, app):
        request_parser = app.app.handlers["service"].services["wms"].request_parser
        try:
            app.app.handlers["service"].services[
                "wms"
            ].request_parser = functools.partial(wms_request, strict=True)

            del self.common_fi_req.params["format"]
            del self.common_fi_req.params["styles"]
            resp = app.get(self.common_fi_req, expect_errors=True)
            xml = resp.lxml
            assert "missing parameters" in xml.xpath("//ServiceException/text()")[0]
            assert validate_with_dtd(xml, "wms/1.1.1/exception_1_1_1.dtd")
        finally:
            app.app.handlers["service"].services["wms"].request_parser = request_parser
            app.app.handlers["service"].request_parser = request_parser

    def test_get_featureinfo_not_queryable(self, app):
        self.common_fi_req.params["query_layers"] = "tms_cache"
        self.common_fi_req.params["exceptions"] = "application/vnd.ogc.se_xml"
        resp = app.get(self.common_fi_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/ServiceException/@code") == []
        assert "tms_cache is not queryable" in xml.xpath("//ServiceException/text()")[0]
        assert validate_with_dtd(xml, "wms/1.1.1/exception_1_1_1.dtd")


class TestWMS110(SysTest):
    config_file = "layer.yaml"

    def setup_method(self):
        # WMSTest.setup_method(self)
        self.common_req = WMS110MapRequest(
            url="/service?", param=dict(service="WMS", version="1.1.0")
        )
        self.common_map_req = WMS110MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.0",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )
        self.common_fi_req = WMS110FeatureInfoRequest(
            url="/service?",
            param=dict(
                x="10",
                y="20",
                width="200",
                height="200",
                layers="wms_cache",
                format="image/png",
                query_layers="wms_cache_110",
                styles="",
                bbox="1000,400,2000,1400",
                srs="EPSG:900913",
            ),
        )

    def test_wms_capabilities(self, app):
        req = WMS110CapabilitiesRequest(url="/service?").copy_with_request_params(
            self.common_req
        )
        resp = app.get(req)
        assert resp.content_type == "application/vnd.ogc.wms_xml"
        xml = resp.lxml
        assert (
            xml.xpath(
                "//GetMap//OnlineResource/@xlink:href",
                namespaces=dict(xlink="http://www.w3.org/1999/xlink"),
            )[0] ==
            "http://localhost/service?"
        )

        llbox = xml.xpath("//Capability/Layer/LatLonBoundingBox")[0]
        # some clients don't like 90deg north/south
        assert float(llbox.attrib["miny"]) == pytest.approx(-70.0)
        assert float(llbox.attrib["maxy"]) == pytest.approx(89.999999)
        assert float(llbox.attrib["minx"]) == pytest.approx(-180.0)
        assert float(llbox.attrib["maxx"]) == pytest.approx(180.0)

        layer_names = set(xml.xpath("//Layer/Layer/Name/text()"))
        expected_names = set(
            [
                "direct_fwd_params",
                "direct",
                "wms_cache",
                "wms_cache_100",
                "wms_cache_130",
                "wms_cache_transparent",
                "wms_merge",
                "tms_cache",
                "tms_fi_cache",
                "wms_cache_multi",
                "wms_cache_link_single",
                "wms_cache_110",
                "watermark_cache",
                "wms_managed_cookies_cache",
            ]
        )
        assert layer_names == expected_names
        assert validate_with_dtd(xml, dtd_name="wms/1.1.0/capabilities_1_1_0.dtd")

    def test_invalid_layer(self, app):
        self.common_map_req.params["layers"] = "invalid"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/@version")[0] == "1.1.0"
        assert (
            xml.xpath("/ServiceExceptionReport/ServiceException/@code")[0] ==
            "LayerNotDefined"
        )
        assert xml.xpath("//ServiceException/text()")[0] == "unknown layer: invalid"
        assert validate_with_dtd(xml, dtd_name="wms/1.1.0/exception_1_1_0.dtd")

    def test_invalid_format(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/@version")[0] == "1.1.0"
        assert (
            xml.xpath("/ServiceExceptionReport/ServiceException/@code")[0] ==
            "InvalidFormat"
        )
        assert (
            xml.xpath("//ServiceException/text()")[0] ==
            "unsupported image format: image/ascii"
        )
        assert validate_with_dtd(xml, dtd_name="wms/1.1.0/exception_1_1_0.dtd")

    def test_invalid_format_img_exception(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_missing_format_img_exception(self, app):
        del self.common_map_req.params["format"]
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_invalid_srs(self, app):
        self.common_map_req.params["srs"] = "EPSG:1234"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/@version")[0] == "1.1.0"
        assert (
            xml.xpath("/ServiceExceptionReport/ServiceException/@code")[0] == "InvalidSRS"
        )
        assert xml.xpath("//ServiceException/text()")[0] == "unsupported srs: EPSG:1234"
        assert validate_with_dtd(xml, dtd_name="wms/1.1.0/exception_1_1_0.dtd")

    def test_get_map_png(self, app, fixture_cache_data):
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        assert Image.open(data).mode == "RGB"

    def test_get_map_jpeg(self, app, fixture_cache_data):
        self.common_map_req.params["format"] = "image/jpeg"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/jpeg"
        assert is_jpeg(BytesIO(resp.body))

    def test_get_map_xml_exception(self, app):
        self.common_map_req.params["bbox"] = "0,0,90,90"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/ServiceException/@code") == []
        assert "No response from URL" in xml.xpath("//ServiceException/text()")[0]
        assert validate_with_dtd(xml, "wms/1.1.0/exception_1_1_0.dtd")

    @pytest.mark.flaky(reruns=5, reruns_delay=2)
    def test_get_map(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"
                resp = app.get(self.common_map_req)
                assert 35000 < int(resp.headers["Content-length"]) < 75000
                assert resp.content_type == "image/png"

        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()

    def test_get_map_110(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.0&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"
                self.common_map_req.params["layers"] = "wms_cache_110"
                resp = app.get(self.common_map_req)
                assert 35000 < int(resp.headers["Content-length"]) < 75000
                assert resp.content_type == "image/png"

        assert cache_dir.join(
            "wms_cache_110_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()

    def test_get_featureinfo(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
                "&VERSION=1.1.0&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=10&Y=20"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_not_queryable(self, app):
        self.common_fi_req.params["query_layers"] = "tms_cache"
        self.common_fi_req.params["exceptions"] = "application/vnd.ogc.se_xml"
        resp = app.get(self.common_fi_req, expect_errors=True)
        assert resp.content_type == "application/vnd.ogc.se_xml"
        xml = resp.lxml
        assert xml.xpath("/ServiceExceptionReport/ServiceException/@code") == []
        assert "tms_cache is not queryable" in xml.xpath("//ServiceException/text()")[0]
        assert validate_with_dtd(xml, "wms/1.1.0/exception_1_1_0.dtd")

    def test_managed_cookies(self, app):
        def assert_no_cookie(req_handler):
            return 'Cookie' not in req_handler.headers

        def assert_cookie(req_handler):
            assert 'Cookie' in req_handler.headers
            cookie_name, cookie_val = req_handler.headers['Cookie'].split(';')[0].split('=')
            assert cookie_name == 'testcookie'
            assert cookie_val == '42'
            return True

        url = (r"/service?LAYERs=layer1&SERVICE=WMS&FORMAT=image%2Fpng"
               "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
               "&VERSION=1.1.1&BBOX=1000.0,400.0,2000.0,1400.0&styles="
               "&WIDTH=200&QUERY_LAYERS=layer1&X=10&Y=20")
        # First response has a Set-Cookie => with managed_cookies=True, mapproxy should send the
        # cookie in the second request
        expected_requests = [
            (
                {'path': url, 'req_assert_function': assert_no_cookie},
                {'body': b'nothing', 'headers': {'Set-Cookie': "testcookie=42"}}
            ),
            (
                {'path': url, 'req_assert_function': assert_cookie},
                {'body': b'nothing'}
            )
        ]
        with mock_httpd(("localhost", 42423), expected_requests):
            self.common_fi_req.params["layers"] = "wms_managed_cookies_cache"
            self.common_fi_req.params["query_layers"] = "wms_managed_cookies_cache"
            resp = app.get(self.common_fi_req)
            assert resp.body == b"nothing"
            resp = app.get(self.common_fi_req)
            assert resp.body == b"nothing"


class TestWMS100(SysTest):
    config_file = "layer.yaml"

    def setup_method(self):
        self.common_req = WMS100MapRequest(url="/service?", param=dict(wmtver="1.0.0"))
        self.common_map_req = WMS100MapRequest(
            url="/service?",
            param=dict(
                wmtver="1.0.0",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache",
                srs="EPSG:4326",
                format="PNG",
                styles="",
                request="GetMap",
            ),
        )
        self.common_fi_req = WMS100FeatureInfoRequest(
            url="/service?",
            param=dict(
                x="10",
                y="20",
                width="200",
                height="200",
                layers="wms_cache_100",
                format="PNG",
                query_layers="wms_cache_100",
                styles="",
                bbox="1000,400,2000,1400",
                srs="EPSG:900913",
            ),
        )

    def test_wms_capabilities(self, app):
        req = WMS100CapabilitiesRequest(url="/service?").copy_with_request_params(
            self.common_req
        )
        resp = app.get(req)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert (
            xml.xpath("/WMT_MS_Capabilities/Service/Title/text()")[0] ==
            u"MapProxy test fixture \u2603"
        )
        layer_names = set(xml.xpath("//Layer/Layer/Name/text()"))
        expected_names = set(
            [
                "direct_fwd_params",
                "direct",
                "wms_cache",
                "wms_cache_100",
                "wms_cache_130",
                "wms_cache_transparent",
                "wms_merge",
                "tms_cache",
                "tms_fi_cache",
                "wms_cache_multi",
                "wms_cache_link_single",
                "wms_cache_110",
                "watermark_cache",
                "wms_managed_cookies_cache",
            ]
        )
        assert layer_names == expected_names
        # TODO srs
        assert validate_with_dtd(xml, dtd_name="wms/1.0.0/capabilities_1_0_0.dtd")

    def test_invalid_layer(self, app):
        self.common_map_req.params["layers"] = "invalid"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert xml.xpath("/WMTException/@version")[0] == "1.0.0"
        assert xml.xpath("//WMTException/text()")[0].strip() == "unknown layer: invalid"

    def test_invalid_format(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert xml.xpath("/WMTException/@version")[0] == "1.0.0"
        assert (
            xml.xpath("//WMTException/text()")[0].strip() ==
            "unsupported image format: ASCII"
        )

    def test_invalid_format_img_exception(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        self.common_map_req.params["exceptions"] = "INIMAGE"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_missing_format_img_exception(self, app):
        del self.common_map_req.params["format"]
        self.common_map_req.params["exceptions"] = "INIMAGE"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_invalid_srs(self, app):
        self.common_map_req.params["srs"] = "EPSG:1234"
        print(self.common_map_req.complete_url)
        resp = app.get(self.common_map_req.complete_url, expect_errors=True)
        xml = resp.lxml
        assert xml.xpath("//WMTException/text()")[0].strip() == "unsupported srs: EPSG:1234"

    def test_get_map_png(self, app, fixture_cache_data):
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        assert Image.open(data).mode == "RGB"

    def test_get_map_png_transparent_paletted(
        self, app, base_config, fixture_cache_data
    ):
        try:
            base_config.image.paletted = True
            self.common_map_req.params["transparent"] = "True"
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"
            data = BytesIO(resp.body)
            assert is_png(data)
            assert Image.open(data).mode == "P"
        finally:
            base_config.image.paletted = False

    def test_get_map_jpeg(self, app, fixture_cache_data):
        self.common_map_req.params["format"] = "image/jpeg"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/jpeg"
        assert is_jpeg(BytesIO(resp.body))

    def test_get_map_xml_exception(self, app):
        self.common_map_req.params["bbox"] = "0,0,90,90"
        resp = app.get(self.common_map_req, expect_errors=True)
        xml = resp.lxml
        assert "No response from URL" in xml.xpath("//WMTException/text()")[0]

    @pytest.mark.flaky(reruns=5, reruns_delay=2)
    def test_get_map(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/png"
        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()

    def test_get_featureinfo(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&FORMAT=image%2FPNG"  # TODO should be PNG only
                "&REQUEST=feature_info&HEIGHT=200&SRS=EPSG%3A900913"
                "&WMTVER=1.0.0&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=10&Y=20"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_not_queryable(self, app):
        self.common_fi_req.params["query_layers"] = "tms_cache"
        self.common_fi_req.params["exceptions"] = "application/vnd.ogc.se_xml"
        resp = app.get(self.common_fi_req, expect_errors=True)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert "tms_cache is not queryable" in xml.xpath("//WMTException/text()")[0]


ns130 = {
    "wms": "http://www.opengis.net/wms",
    "ogc": "http://www.opengis.net/ogc",
    "sld": "http://www.opengis.net/sld",
    "xlink": "http://www.w3.org/1999/xlink",
}


def assert_xpath(xml, xpath, expected, namespaces=None):
    assert xml.xpath(xpath, namespaces=namespaces)[0] == expected


assert_xpath_wms130 = functools.partial(assert_xpath, namespaces=ns130)


class TestWMS130(SysTest):
    config_file = "layer.yaml"

    def setup_method(self):
        self.common_req = WMS130MapRequest(
            url="/service?", param=dict(service="WMS", version="1.3.0")
        )
        self.common_map_req = WMS130MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.3.0",
                bbox="0,-180,80,0",
                width="200",
                height="200",
                layers="wms_cache",
                crs="EPSG:4326",
                format="image/png",
                styles="",
                request="GetMap",
            ),
        )
        self.common_fi_req = WMS130FeatureInfoRequest(
            url="/service?",
            param=dict(
                i="10",
                j="20",
                width="200",
                height="200",
                layers="wms_cache_130",
                format="image/png",
                query_layers="wms_cache_130",
                styles="",
                bbox="1000,400,2000,1400",
                crs="EPSG:900913",
            ),
        )

    def test_wms_capabilities(self, app):
        req = WMS130CapabilitiesRequest(url="/service?").copy_with_request_params(
            self.common_req
        )
        resp = app.get(req)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert_xpath_wms130(
            xml,
            "/wms:WMS_Capabilities/wms:Service/wms:Title/text()",
            u"MapProxy test fixture \u2603",
        )

        # test for extended layer metadata
        assert_xpath_wms130(
            xml,
            "/wms:WMS_Capabilities/wms:Capability/wms:Layer/wms:Layer/wms:Attribution/wms:Title/text()",
            u"My attribution title",
        )

        layer_names = set(
            xml.xpath("//wms:Layer/wms:Layer/wms:Name/text()", namespaces=ns130)
        )
        expected_names = set(
            [
                "direct_fwd_params",
                "direct",
                "wms_cache",
                "wms_cache_100",
                "wms_cache_130",
                "wms_cache_transparent",
                "wms_merge",
                "tms_cache",
                "tms_fi_cache",
                "wms_cache_multi",
                "wms_cache_link_single",
                "wms_cache_110",
                "watermark_cache",
                "wms_managed_cookies_cache",
            ]
        )
        assert layer_names == expected_names
        assert is_130_capa(xml)

    def test_invalid_layer(self, app):
        self.common_map_req.params["layers"] = "invalid"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert_xpath_wms130(xml, "/ogc:ServiceExceptionReport/@version", "1.3.0")
        assert_xpath_wms130(
            xml,
            "/ogc:ServiceExceptionReport/ogc:ServiceException/@code",
            "LayerNotDefined",
        )
        assert_xpath_wms130(xml, "//ogc:ServiceException/text()", "unknown layer: invalid")
        assert validate_with_xsd(xml, xsd_name="wms/1.3.0/exceptions_1_3_0.xsd")

    def test_invalid_format(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert_xpath_wms130(xml, "/ogc:ServiceExceptionReport/@version", "1.3.0")
        assert_xpath_wms130(
            xml,
            "/ogc:ServiceExceptionReport/ogc:ServiceException/@code",
            "InvalidFormat",
        )
        assert_xpath_wms130(
            xml,
            "//ogc:ServiceException/text()",
            "unsupported image format: image/ascii",
        )
        assert validate_with_xsd(xml, xsd_name="wms/1.3.0/exceptions_1_3_0.xsd")

    def test_invalid_format_img_exception(self, app):
        self.common_map_req.params["format"] = "image/ascii"
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_missing_format_img_exception(self, app):
        del self.common_map_req.params["format"]
        self.common_map_req.params["exceptions"] = "application/vnd.ogc.se_inimage"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        assert is_png(BytesIO(resp.body))

    def test_invalid_srs(self, app):
        self.common_map_req.params["srs"] = "EPSG:1234"
        self.common_map_req.params["exceptions"] = "text/xml"

        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert_xpath_wms130(
            xml, "/ogc:ServiceExceptionReport/ogc:ServiceException/@code", "InvalidCRS"
        )
        assert_xpath_wms130(
            xml, "//ogc:ServiceException/text()", "unsupported crs: EPSG:1234"
        )
        assert validate_with_xsd(xml, xsd_name="wms/1.3.0/exceptions_1_3_0.xsd")

    def test_get_map_png(self, app, fixture_cache_data):
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/png"
        data = BytesIO(resp.body)
        assert is_png(data)
        assert Image.open(data).mode == "RGB"

    def test_get_map_jpeg(self, app, fixture_cache_data):
        self.common_map_req.params["format"] = "image/jpeg"
        resp = app.get(self.common_map_req)
        assert resp.content_type == "image/jpeg"
        assert is_jpeg(BytesIO(resp.body))

    def test_get_map_xml_exception(self, app):
        self.common_map_req.params["bbox"] = "0,0,90,90"
        resp = app.get(self.common_map_req, expect_errors=True)
        assert resp.content_type == "text/xml"
        xml = resp.lxml
        assert (
            xml.xpath(
                "/ogc:ServiceExceptionReport/ogc:ServiceException/@code",
                namespaces=ns130,
            ) ==
            []
        )
        assert (
            "No response from URL"
            in xml.xpath("//ogc:ServiceException/text()", namespaces=ns130)[0]
        )
        assert validate_with_xsd(xml, xsd_name="wms/1.3.0/exceptions_1_3_0.xsd")

    @pytest.mark.flaky(reruns=5, reruns_delay=2)
    def test_get_map(self, app, cache_dir):
        with tmp_image((256, 256), format="jpeg") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"  # internal axis-order
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/png"

        assert cache_dir.join(
            "wms_cache_EPSG900913/01/000/000/001/000/000/001.jpeg"
        ).check()

    def test_get_featureinfo(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&CRS=EPSG%3A900913"
                "&VERSION=1.3.0&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&I=10&J=20"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"

    def test_get_featureinfo_111(self, app):
        expected_req = (
            {
                "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fpng"
                "&REQUEST=GetFeatureInfo&HEIGHT=200&SRS=EPSG%3A900913"
                "&VERSION=1.1.1&BBOX=1000.0,400.0,2000.0,1400.0&styles="
                "&WIDTH=200&QUERY_LAYERS=foo,bar&X=10&Y=20"
            },
            {"body": b"info", "headers": {"content-type": "text/plain"}},
        )
        with mock_httpd(("localhost", 42423), [expected_req]):
            self.common_fi_req.params["layers"] = "wms_cache"
            self.common_fi_req.params["query_layers"] = "wms_cache"
            resp = app.get(self.common_fi_req)
            assert resp.content_type == "text/plain"
            assert resp.body == b"info"


@pytest.mark.skipif(sys.platform == "win32", reason="not supported on Windows")
class TestWMSLinkSingleColorImages(SysTest):
    config_file = "layer.yaml"

    def setup_method(self):
        self.common_map_req = WMS111MapRequest(
            url="/service?",
            param=dict(
                service="WMS",
                version="1.1.1",
                bbox="-180,0,0,80",
                width="200",
                height="200",
                layers="wms_cache_link_single",
                srs="EPSG:4326",
                format="image/jpeg",
                styles="",
                request="GetMap",
            ),
        )

    @pytest.mark.flaky(reruns=5, reruns_delay=2)
    def test_get_map(self, app, cache_dir):
        link_name = "wms_cache_link_single_EPSG900913/01/000/000/001/000/000/001.png"
        real_name = "wms_cache_link_single_EPSG900913/single_color_tiles/fe00a0.png"
        with tmp_image((256, 256), format="jpeg", color="#fe00a0") as img:
            expected_req = (
                {
                    "path": r"/service?LAYERs=foo,bar&SERVICE=WMS&FORMAT=image%2Fjpeg"
                    "&REQUEST=GetMap&HEIGHT=256&SRS=EPSG%3A900913&styles="
                    "&VERSION=1.1.1&BBOX=0.0,0.0,20037508.3428,20037508.3428"
                    "&WIDTH=256"
                },
                {"body": img.read(), "headers": {"content-type": "image/jpeg"}},
            )
            with mock_httpd(
                ("localhost", 42423), [expected_req], bbox_aware_query_comparator=True
            ):
                self.common_map_req.params["bbox"] = "0,0,180,90"
                resp = app.get(self.common_map_req)
                assert resp.content_type == "image/jpeg"

            single_loc = cache_dir.join(real_name)
            tile_loc = cache_dir.join(link_name)
            assert single_loc.check()
            assert tile_loc.check(link=True)

            self.common_map_req.params["format"] = "image/png"
            resp = app.get(self.common_map_req)
            assert resp.content_type == "image/png"


def assert_almost_equal_bbox(bbox1, bbox2, rel=0.01):
    assert bbox1 == pytest.approx(bbox2, rel=rel)


def is_100_capa(xml):
    return validate_with_dtd(xml, dtd_name="wms/1.0.0/capabilities_1_0_0.dtd")


def is_110_capa(xml):
    return validate_with_dtd(xml, dtd_name="wms/1.1.0/capabilities_1_1_0.dtd")


def is_111_exception(xml, msg=None, code=None, re_msg=None):
    assert xml.xpath("/ServiceExceptionReport/@version")[0] == "1.1.1"
    if msg:
        assert xml.xpath("//ServiceException/text()")[0] == msg
    if re_msg:
        exception_msg = xml.xpath("//ServiceException/text()")[0]
        assert re.findall(re_msg, exception_msg, re.I), "'%r' does not match '%s'" % (
            re_msg,
            exception_msg,
        )
    if code is not None:
        assert xml.xpath("/ServiceExceptionReport/ServiceException/@code")[0] == code
    assert validate_with_dtd(xml, "wms/1.1.1/exception_1_1_1.dtd")


def is_111_capa(xml):
    return validate_with_dtd(xml, dtd_name="wms/1.1.1/WMS_MS_Capabilities.dtd")


def is_130_capa(xml):
    return validate_with_xsd(xml, xsd_name="wms/1.3.0/capabilities_1_3_0.xsd")



# This file is part of the MapProxy project.
# Copyright (C) 2022 Even Rouault
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import pytest

from mapproxy.test.system import SysTest


@pytest.fixture(scope="module")
def config_file():
    return "demo.yaml"


class TestDemo(SysTest):

    def test_basic(self, app):
        resp = app.get("/demo/", status=200)
        assert resp.content_type == "text/html"
        assert 'href="../service?REQUEST=GetCapabilities&SERVICE=WMS"' in resp
        assert 'href="../service?REQUEST=GetCapabilities&SERVICE=WMS&tiled=true"' in resp
        assert 'href="../demo/?wmts_layer=wms_cache&format=jpeg&srs=EPSG%3A900913"' in resp
        assert 'href="../demo/?tms_layer=wms_cache&format=jpeg&srs=EPSG%3A900913"' in resp

    def test_previewmap(self, app):
        resp = app.get("/demo/?srs=EPSG%3A3857&format=image%2Fpng&wms_layer=wms_cache", status=200)
        assert resp.content_type == "text/html"
        assert '<h2>Layer Preview - wms_cache</h2>' in resp

    def test_layers_sorted_by_name(self, app):
        resp = app.get("/demo/", status=200)

        patternTable = re.compile(r'<table class="code">.*?</table>', re.DOTALL)
        patternWMS = r'<td\s+rowspan="2">([\w\d-]+)</td>'
        patternWMTS_TBS = r'<td\s+rowspan="1">([\w\d-]+)</td>'
        tables = patternTable.findall(resp.text)

        layersWMS = re.findall(patternWMS, resp.text, re.IGNORECASE)
        layersWMTS = re.findall(patternWMTS_TBS, tables[1], re.IGNORECASE)
        layersTBS = re.findall(patternWMTS_TBS, tables[2], re.IGNORECASE)

        assert layersWMS == sorted(layersWMS)
        assert layersWMTS == sorted(layersWMTS)
        assert layersTBS == sorted(layersTBS)



# This file is part of the MapProxy project.
# Copyright (C) 2010 Omniscale <http://omniscale.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
(Meta-)Tile grids (data and calculations).
"""
from __future__ import division
import math

from mapproxy.srs import SRS, get_epsg_num, merge_bbox, bbox_equals
from mapproxy.util.collections import ImmutableDictList

geodetic_epsg_codes = [4326]


class GridError(Exception):
    pass


class NoTiles(GridError):
    pass


def get_resolution(bbox, size):
    """
    Calculate the highest resolution needed to draw the bbox
    into an image with given size.

    >>> get_resolution((-180,-90,180,90), (256, 256))
    0.703125

    :returns: the resolution
    :rtype: float
    """
    w = abs(bbox[0] - bbox[2])
    h = abs(bbox[1] - bbox[3])
    return min(w/size[0], h/size[1])


def tile_grid_for_epsg(epsg, bbox=None, tile_size=(256, 256), res=None):
    """
    Create a tile grid that matches the given epsg code:

    :param epsg: the epsg code
    :type epsg: 'EPSG:0000', '0000' or 0000
    :param bbox: the bbox of the grid
    :param tile_size: the size of each tile
    :param res: a list with all resolutions
    """
    epsg = get_epsg_num(epsg)
    if epsg in geodetic_epsg_codes:
        return TileGrid(epsg, is_geodetic=True, bbox=bbox, tile_size=tile_size, res=res)
    return TileGrid(epsg, bbox=bbox, tile_size=tile_size, res=res)


# defer loading of default bbox since custom proj settings
# are not loaded on import time
class _default_bboxs(object):
    _defaults = {
        4326: (-180, -90, 180, 90),
    }
    for epsg_num in (900913, 3857, 102100, 102113):
        _defaults[epsg_num] = (-20037508.342789244,
                               -20037508.342789244,
                               20037508.342789244,
                               20037508.342789244)
    defaults = None

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        if self.defaults is None:
            defaults = {}
            for epsg, bbox in self._defaults.items():
                defaults[SRS(epsg)] = bbox
            self.defaults = defaults
        return self.defaults[key]


default_bboxs = _default_bboxs()


def tile_grid(srs=None, bbox=None, bbox_srs=None, tile_size=(256, 256),
              res=None, res_factor=2.0, threshold_res=None,
              num_levels=None, min_res=None, max_res=None,
              stretch_factor=1.15, max_shrink_factor=4.0,
              align_with=None, origin='ll', name=None
              ):
    """
    This function creates a new TileGrid.
    """
    if srs is None:
        srs = 'EPSG:900913'
    srs = SRS(srs)

    if not bbox:
        bbox = default_bboxs.get(srs)
        if not bbox:
            raise ValueError('need a bbox for grid with %s' % srs)

    bbox = grid_bbox(bbox, srs=srs, bbox_srs=bbox_srs)

    if res:
        if isinstance(res, list):
            if isinstance(res[0], (tuple, list)):
                # named resolutions
                res = sorted(res, key=lambda x: x[1], reverse=True)
            else:
                res = sorted(res, reverse=True)
            assert min_res is None
            assert max_res is None
            assert align_with is None
        else:
            raise ValueError("res is not a list, use res_factor for float values")

    elif align_with is not None:
        res = aligned_resolutions(min_res, max_res, res_factor, num_levels, bbox, tile_size,
                                  align_with)
    else:
        res = resolutions(min_res, max_res, res_factor, num_levels, bbox, tile_size)

    origin = origin_from_string(origin)

    return TileGrid(srs, bbox=bbox, tile_size=tile_size, res=res, threshold_res=threshold_res,
                    stretch_factor=stretch_factor, max_shrink_factor=max_shrink_factor,
                    origin=origin, name=name)


ORIGIN_UL = 'ul'
ORIGIN_LL = 'll'


def origin_from_string(origin):
    if origin is None:
        origin = ORIGIN_LL
    elif origin.lower() in ('ll', 'sw'):
        origin = ORIGIN_LL
    elif origin.lower() in ('ul', 'nw'):
        origin = ORIGIN_UL
    else:
        raise ValueError("unknown origin value '%s'" % origin)
    return origin


def aligned_resolutions(min_res=None, max_res=None, res_factor=2.0, num_levels=None,
                        bbox=None, tile_size=(256, 256), align_with=None):

    alinged_res = align_with.resolutions
    res = list(alinged_res)

    if not min_res:
        width = bbox[2] - bbox[0]
        height = bbox[3] - bbox[1]
        min_res = max(width/tile_size[0], height/tile_size[1])

    res = [r for r in res if r <= min_res]

    if max_res:
        res = [r for r in res if r >= max_res]

    if num_levels:
        res = res[:num_levels]

    factor_calculated = res[0]/res[1]
    if res_factor == 'sqrt2' and round(factor_calculated, 8) != round(math.sqrt(2), 8):
        if round(factor_calculated, 8) == 2.0:
            new_res = []
            for r in res:
                new_res.append(r)
                new_res.append(r/math.sqrt(2))
            res = new_res
    elif res_factor == 2.0 and round(factor_calculated, 8) != round(2.0, 8):
        if round(factor_calculated, 8) == round(math.sqrt(2), 8):
            res = res[::2]
    return res


def resolutions(min_res=None, max_res=None, res_factor=2.0, num_levels=None,
                bbox=None, tile_size=(256, 256)):
    if res_factor == 'sqrt2':
        res_factor = math.sqrt(2)

    res = []
    if not min_res:
        width = bbox[2] - bbox[0]
        height = bbox[3] - bbox[1]
        min_res = max(width/tile_size[0], height/tile_size[1])

    if max_res:
        if num_levels:
            res_step = (math.log10(min_res) - math.log10(max_res)) / (num_levels-1)
            res = [10**(math.log10(min_res) - res_step*i) for i in range(num_levels)]
        else:
            res = [min_res]
            while True:
                next_res = res[-1]/res_factor
                if max_res >= next_res:
                    break
                res.append(next_res)
    else:
        if not num_levels:
            num_levels = 20 if res_factor != math.sqrt(2) else 40
        res = [min_res]
        while len(res) < num_levels:
            res.append(res[-1]/res_factor)

    return res


def grid_bbox(bbox, bbox_srs, srs):
    bbox = bbox_tuple(bbox)
    if bbox_srs:
        bbox = SRS(bbox_srs).transform_bbox_to(srs, bbox)
    return bbox


def bbox_tuple(bbox):
    """
    >>> bbox_tuple('20,-30,40,-10')
    (20.0, -30.0, 40.0, -10.0)
    >>> bbox_tuple([20,-30,40,-10])
    (20.0, -30.0, 40.0, -10.0)

    """
    if isinstance(bbox, str):
        bbox = bbox.split(',')
    bbox = tuple(map(float, bbox))
    return bbox


def bbox_width(bbox):
    return bbox[2] - bbox[0]


def bbox_height(bbox):
    return bbox[3] - bbox[1]


def bbox_size(bbox):
    return bbox_width(bbox), bbox_height(bbox)


class NamedGridList(ImmutableDictList):
    def __init__(self, items):
        tmp = []
        for i, value in enumerate(items):
            if isinstance(value, (tuple, list)):
                name, value = value
            else:
                name = str('%02d' % i)
            tmp.append((name, value))
        ImmutableDictList.__init__(self, tmp)


class TileGrid(object):
    """
    This class represents a regular tile grid. The first level (0) contains a single
    tile, the origin is bottom-left.

    :ivar levels: the number of levels
    :ivar tile_size: the size of each tile in pixel
    :type tile_size: ``int(with), int(height)``
    :ivar srs: the srs of the grid
    :type srs: `SRS`
    :ivar bbox: the bbox of the grid, tiles may overlap this bbox
    """

    spheroid_a = 6378137.0  # for 900913
    flipped_y_axis = False

    def __init__(self, srs=900913, bbox=None, tile_size=(256, 256), res=None,
                 threshold_res=None, is_geodetic=False, levels=None,
                 stretch_factor=1.15, max_shrink_factor=4.0, origin='ll',
                 name=None):
        """
        :param stretch_factor: allow images to be scaled up by this factor
            before the next level will be selected
        :param max_shrink_factor: allow images to be scaled down by this
            factor before NoTiles is raised

        >>> grid = TileGrid(srs=900913)
        >>> [round(x, 2) for x in grid.bbox]
        [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
        """
        if isinstance(srs, (int, str)):
            srs = SRS(srs)
        self.srs = srs
        self.tile_size = tile_size
        self.origin = origin_from_string(origin)
        self.name = name

        if self.origin == 'ul':
            self.flipped_y_axis = True

        self.is_geodetic = is_geodetic

        self.stretch_factor = stretch_factor
        self.max_shrink_factor = max_shrink_factor

        if levels is None:
            self.levels = 20
        else:
            self.levels = levels

        if bbox is None:
            bbox = self._calc_bbox()
        self.bbox = bbox

        factor = None

        if res is None:
            factor = 2.0
            res = self._calc_res(factor=factor)
        elif res == 'sqrt2':
            if levels is None:
                self.levels = 40
            factor = math.sqrt(2)
            res = self._calc_res(factor=factor)
        elif is_float(res):
            factor = float(res)
            res = self._calc_res(factor=factor)

        self.levels = len(res)
        self.resolutions = NamedGridList(res)

        self.threshold_res = None
        if threshold_res:
            self.threshold_res = sorted(threshold_res)

        self.grid_sizes = self._calc_grids()

    def _calc_grids(self):
        width = self.bbox[2] - self.bbox[0]
        height = self.bbox[3] - self.bbox[1]
        grids = []
        for idx, res in self.resolutions.iteritems():
            x = max(math.ceil(width // res / self.tile_size[0]), 1)
            y = max(math.ceil(height // res / self.tile_size[1]), 1)
            grids.append((idx, (int(x), int(y))))
        return NamedGridList(grids)

    def _calc_bbox(self):
        if self.is_geodetic:
            return (-180.0, -90.0, 180.0, 90.0)
        else:
            circum = 2 * math.pi * self.spheroid_a
            offset = circum / 2.0
            return (-offset, -offset, offset, offset)

    def _calc_res(self, factor=None):
        width = self.bbox[2] - self.bbox[0]
        height = self.bbox[3] - self.bbox[1]
        initial_res = max(width/self.tile_size[0], height/self.tile_size[1])
        if factor is None:
            return pyramid_res_level(initial_res, levels=self.levels)
        else:
            return pyramid_res_level(initial_res, factor, levels=self.levels)

    def resolution(self, level):
        """
        Returns the resolution of the `level` in units/pixel.

        :param level: the zoom level index (zero is top)

        >>> grid = TileGrid(SRS(900913))
        >>> '%.5f' % grid.resolution(0)
        '156543.03393'
        >>> '%.5f' % grid.resolution(1)
        '78271.51696'
        >>> '%.5f' % grid.resolution(4)
        '9783.93962'
        """
        return self.resolutions[level]

    def closest_level(self, res):
        """
        Returns the level index that offers the required resolution.

        :param res: the required resolution
        :returns: the level with the requested or higher resolution

        >>> grid = TileGrid(SRS(900913))
        >>> grid.stretch_factor = 1.1
        >>> l1_res = grid.resolution(1)
        >>> [grid.closest_level(x) for x in (320000.0, 160000.0, l1_res+50, l1_res, \
                                             l1_res-50, l1_res*0.91, l1_res*0.89, 8000.0)]
        [0, 0, 1, 1, 1, 1, 2, 5]
        """
        prev_l_res = self.resolutions[0]
        threshold = None
        thresholds = []
        if self.threshold_res:
            thresholds = self.threshold_res[:]
            threshold = thresholds.pop()
            # skip thresholds above first res
            while threshold > prev_l_res and thresholds:
                threshold = thresholds.pop()

        threshold_result = None
        for level, l_res in enumerate(self.resolutions):
            if threshold and prev_l_res > threshold >= l_res:
                if res > threshold:
                    return level-1
                elif res >= l_res:
                    return level
                threshold = thresholds.pop() if thresholds else None

            if threshold_result is not None:
                # Use previous level that was within stretch_factor,
                # but only if this level res is smaller then res.
                # This fixes selection for resolutions that are closer together then stretch_factor.
                #
                if l_res < res:
                    return threshold_result

            if l_res <= res*self.stretch_factor:
                # l_res within stretch_factor
                # remember this level, check for thresholds or better res in next loop
                threshold_result = level
            prev_l_res = l_res
        return level

    def tile(self, x, y, level):
        """
        Returns the tile id for the given point.

        >>> grid = TileGrid(SRS(900913))
        >>> grid.tile(1000, 1000, 0)
        (0, 0, 0)
        >>> grid.tile(1000, 1000, 1)
        (1, 1, 1)
        >>> grid = TileGrid(SRS(900913), tile_size=(512, 512))
        >>> grid.tile(1000, 1000, 2)
        (2, 2, 2)
        """
        res = self.resolution(level)
        x = x - self.bbox[0]
        if self.flipped_y_axis:
            y = self.bbox[3] - y
        else:
            y = y - self.bbox[1]
        tile_x = x/float(res*self.tile_size[0])
        tile_y = y/float(res*self.tile_size[1])
        return (int(math.floor(tile_x)), int(math.floor(tile_y)), level)

    def flip_tile_coord(self, tile_coord):
        """
        Flip the tile coord on the y-axis. (Switch between bottom-left and top-left
        origin.)

        >>> grid = TileGrid(SRS(900913))
        >>> grid.flip_tile_coord((0, 1, 1))
        (0, 0, 1)
        >>> grid.flip_tile_coord((1, 3, 2))
        (1, 0, 2)
        """
        (x, y, z) = tile_coord
        return (x, self.grid_sizes[z][1]-1-y, z)

    def supports_access_with_origin(self, origin):
        if origin_from_string(origin) == self.origin:
            return True

        # check for each level if the top and bottom coordinates of the tiles
        # match the bbox of the grid. only in this case we can flip y-axis
        # without any issues

        # allow for some rounding errors in the _tiles_bbox calculations
        delta = max(abs(self.bbox[1]), abs(self.bbox[3])) / 1e12

        for level, grid_size in enumerate(self.grid_sizes):
            level_bbox = self._tiles_bbox([(0, 0, level),
                                           (grid_size[0] - 1, grid_size[1] - 1, level)])

            if abs(self.bbox[1] - level_bbox[1]) > delta or abs(self.bbox[3] - level_bbox[3]) > delta:
                return False
        return True

    def origin_tile(self, level, origin):
        assert self.supports_access_with_origin(origin), 'tile origins are incompatible'
        tile = (0, 0, level)

        if origin_from_string(origin) == self.origin:
            return tile

        return self.flip_tile_coord(tile)

    def get_affected_tiles(self, bbox, size, req_srs=None):
        """
        Get a list with all affected tiles for a bbox and output size.

        :returns: the bbox, the size and a list with tile coordinates, sorted row-wise
        :rtype: ``bbox, (xs, yz), [(x, y, z), ...]``

        >>> grid = TileGrid()
        >>> bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        >>> tile_size = (256, 256)
        >>> grid.get_affected_tiles(bbox, tile_size)
        ... #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        ((-20037508.342789244, -20037508.342789244,\
          20037508.342789244, 20037508.342789244), (1, 1),\
          <generator object ...>)
        """
        src_bbox, level = self.get_affected_bbox_and_level(bbox, size, req_srs=req_srs)
        return self.get_affected_level_tiles(src_bbox, level)

    def get_affected_bbox_and_level(self, bbox, size, req_srs=None):
        if req_srs and req_srs != self.srs:
            src_bbox = req_srs.transform_bbox_to(self.srs, bbox)
        else:
            src_bbox = bbox

        if not bbox_intersects(self.bbox, src_bbox):
            raise NoTiles()

        res = get_resolution(src_bbox, size)
        level = self.closest_level(res)

        if res > self.resolutions[0]*self.max_shrink_factor:
            raise NoTiles()

        return src_bbox, level

    def get_affected_level_tiles(self, bbox, level):
        """
        Get a list with all affected tiles for a `bbox` in the given `level`.
        :returns: the bbox, the size and a list with tile coordinates, sorted row-wise
        :rtype: ``bbox, (xs, yz), [(x, y, z), ...]``

        >>> grid = TileGrid()
        >>> bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        >>> grid.get_affected_level_tiles(bbox, 0)
        ... #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        ((-20037508.342789244, -20037508.342789244,\
          20037508.342789244, 20037508.342789244), (1, 1),\
          <generator object ...>)
        """
        # remove 1/10 of a pixel so we don't get a tiles we only touch
        delta = self.resolutions[level] / 10.0
        x0, y0, _ = self.tile(bbox[0]+delta, bbox[1]+delta, level)
        x1, y1, _ = self.tile(bbox[2]-delta, bbox[3]-delta, level)
        try:
            return self._tile_iter(x0, y0, x1, y1, level)
        except IndexError:
            raise GridError('Invalid BBOX')

    def _tile_iter(self, x0, y0, x1, y1, level):
        xs = list(range(x0, x1+1))
        if self.flipped_y_axis:
            y0, y1 = y1, y0
            ys = list(range(y0, y1+1))
        else:
            ys = list(range(y1, y0-1, -1))

        ll = (xs[0], ys[-1], level)
        ur = (xs[-1], ys[0], level)

        abbox = self._tiles_bbox([ll, ur])
        return (abbox, (len(xs), len(ys)),
                _create_tile_list(xs, ys, level, self.grid_sizes[level]))

    def _tiles_bbox(self, tiles):
        """
        Returns the bbox of multiple tiles.
        The tiles should be ordered row-wise, bottom-up.

        :param tiles: ordered list of tiles
        :returns: the bbox of all tiles
        """
        ll_bbox = self.tile_bbox(tiles[0])
        ur_bbox = self.tile_bbox(tiles[-1])
        return merge_bbox(ll_bbox, ur_bbox)

    def tile_bbox(self, tile_coord, limit=False):
        """
        Returns the bbox of the given tile.

        >>> grid = TileGrid(SRS(900913))
        >>> [round(x, 2) for x in grid.tile_bbox((0, 0, 0))]
        [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
        >>> [round(x, 2) for x in grid.tile_bbox((1, 1, 1))]
        [0.0, 0.0, 20037508.34, 20037508.34]
        """
        x, y, z = tile_coord
        res = self.resolution(z)

        x0 = self.bbox[0] + round(x * res * self.tile_size[0], 12)
        x1 = x0 + round(res * self.tile_size[0], 12)

        if self.flipped_y_axis:
            y1 = self.bbox[3] - round(y * res * self.tile_size[1], 12)
            y0 = y1 - round(res * self.tile_size[1], 12)
        else:
            y0 = self.bbox[1] + round(y * res * self.tile_size[1], 12)
            y1 = y0 + round(res * self.tile_size[1], 12)

        if limit:
            return (
                max(x0, self.bbox[0]),
                max(y0, self.bbox[1]),
                min(x1, self.bbox[2]),
                min(y1, self.bbox[3])
            )

        return x0, y0, x1, y1

    def limit_tile(self, tile_coord):
        """
        Check if the `tile_coord` is in the grid.

        :returns: the `tile_coord` if it is within the ``grid``,
                  otherwise ``None``.

        >>> grid = TileGrid(SRS(900913))
        >>> grid.limit_tile((-1, 0, 2)) is None
        True
        >>> grid.limit_tile((1, 2, 1)) is None
        True
        >>> grid.limit_tile((1, 2, 2))
        (1, 2, 2)
        """
        x, y, z = tile_coord
        if isinstance(z, str):
            if z not in self.grid_sizes:
                return None
        elif z < 0 or z >= self.levels:
            return None
        grid = self.grid_sizes[z]
        if x < 0 or y < 0 or x >= grid[0] or y >= grid[1]:
            return None
        return x, y, z

    def __repr__(self):
        return '%s(%r, (%.4f, %.4f, %.4f, %.4f),...)' % (
            self.__class__.__name__, self.srs, self.bbox[0], self.bbox[1], self.bbox[2], self.bbox[3])

    def is_subset_of(self, other):
        """
        Returns ``True`` if every tile in `self` is present in `other`.
        Tile coordinates might differ and `other` may contain more
        tiles (more levels, larger bbox).
        """
        if self.srs != other.srs:
            return False

        if self.tile_size != other.tile_size:
            return False

        # check if all level tiles from self align with (affected)
        # tiles from other
        for self_level, self_level_res in self.resolutions.iteritems():
            level_size = (
                self.grid_sizes[self_level][0] * self.tile_size[0],
                self.grid_sizes[self_level][1] * self.tile_size[1]
            )
            level_bbox = self._tiles_bbox([
                (0, 0, self_level),
                (self.grid_sizes[self_level][0] - 1, self.grid_sizes[self_level][1] - 1, self_level)
            ])

            try:
                bbox, level = other.get_affected_bbox_and_level(level_bbox, level_size)
            except NoTiles:
                return False
            try:
                bbox, grid_size, tiles = other.get_affected_level_tiles(level_bbox, level)
            except GridError:
                return False

            if other.resolution(level) != self_level_res:
                return False
            if not bbox_equals(bbox, level_bbox):
                return False

        return True


def _create_tile_list(xs, ys, level, grid_size):
    """
    Returns an iterator tile_coords for the given tile ranges (`xs` and `ys`).
    If the one tile_coord is negative or out of the `grid_size` bound,
    the coord is None.
    """
    x_limit = grid_size[0]
    y_limit = grid_size[1]
    for y in ys:
        for x in xs:
            if x < 0 or y < 0 or x >= x_limit or y >= y_limit:
                yield None
            else:
                yield x, y, level


def is_float(x):
    try:
        float(x)
        return True
    except TypeError:
        return False


def pyramid_res_level(initial_res, factor=2.0, levels=20):
    """
    Return resolutions of an image pyramid.

    :param initial_res: the resolution of the top level (0)
    :param factor: the factor between each level, for tms access 2
    :param levels: number of resolutions to generate

    >>> list(pyramid_res_level(10000, levels=5))
    [10000.0, 5000.0, 2500.0, 1250.0, 625.0]
    >>> [round(x, 4) for x in
    ...     pyramid_res_level(10000, factor=1/0.75, levels=5)]
    [10000.0, 7500.0, 5625.0, 4218.75, 3164.0625]
    """
    return [initial_res/factor**n for n in range(levels)]


class MetaGrid(object):
    """
    This class contains methods to calculate bbox, etc. of metatiles.

    :param grid: the grid to use for the metatiles
    :param meta_size: the number of tiles a metatile consist
    :type meta_size: ``(x_size, y_size)``
    :param meta_buffer: the buffer size in pixel that is added to each metatile.
        the number is added to all four borders.
        this buffer may improve the handling of lables overlapping (meta)tile borders.
    :type meta_buffer: pixel
    """

    def __init__(self, grid, meta_size, meta_buffer=0):
        self.grid = grid
        self.meta_size = meta_size or 0
        self.meta_buffer = meta_buffer

    def _meta_bbox(self, tile_coord=None, tiles=None, limit_to_bbox=True):
        """
        Returns the bbox of the metatile that contains `tile_coord`.

        :type tile_coord: ``(x, y, z)``

        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> [round(x, 2) for x in mgrid._meta_bbox((0, 0, 2))[0]]
        [-20037508.34, -20037508.34, 0.0, 0.0]
        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> [round(x, 2) for x in mgrid._meta_bbox((0, 0, 0))[0]]
        [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
        """
        if tiles:
            assert tile_coord is None
            level = tiles[0][2]
            bbox = self.grid._tiles_bbox(tiles)
        else:
            level = tile_coord[2]
            bbox = self.unbuffered_meta_bbox(tile_coord)
        return self._buffered_bbox(bbox, level, limit_to_bbox)

    def unbuffered_meta_bbox(self, tile_coord):
        x, y, z = tile_coord

        meta_size = self._meta_size(z)

        return self.grid._tiles_bbox([(tile_coord),
                                      (x+meta_size[0]-1, y+meta_size[1]-1, z)])

    def _buffered_bbox(self, bbox, level, limit_to_grid_bbox=True):
        minx, miny, maxx, maxy = bbox

        buffers = (0, 0, 0, 0)
        if self.meta_buffer > 0:
            res = self.grid.resolution(level)
            minx -= self.meta_buffer * res
            miny -= self.meta_buffer * res
            maxx += self.meta_buffer * res
            maxy += self.meta_buffer * res
            buffers = [self.meta_buffer, self.meta_buffer, self.meta_buffer, self.meta_buffer]

            if limit_to_grid_bbox:
                if self.grid.bbox[0] > minx:
                    delta = self.grid.bbox[0] - minx
                    buffers[0] = buffers[0] - int(round(delta / res, 5))
                    minx = self.grid.bbox[0]
                if self.grid.bbox[1] > miny:
                    delta = self.grid.bbox[1] - miny
                    buffers[1] = buffers[1] - int(round(delta / res, 5))
                    miny = self.grid.bbox[1]
                if self.grid.bbox[2] < maxx:
                    delta = maxx - self.grid.bbox[2]
                    buffers[2] = buffers[2] - int(round(delta / res, 5))
                    maxx = self.grid.bbox[2]
                if self.grid.bbox[3] < maxy:
                    delta = maxy - self.grid.bbox[3]
                    buffers[3] = buffers[3] - int(round(delta / res, 5))
                    maxy = self.grid.bbox[3]
        return (minx, miny, maxx, maxy), tuple(buffers)

    def meta_tile(self, tile_coord):
        """
        Returns the meta tile for `tile_coord`.
        """
        tile_coord = self.main_tile(tile_coord)
        level = tile_coord[2]
        bbox, buffers = self._meta_bbox(tile_coord)
        grid_size = self._meta_size(level)
        size = self._size_from_buffered_bbox(bbox, level)

        tile_patterns = self._tiles_pattern(tile=tile_coord, grid_size=grid_size, buffers=buffers)

        return MetaTile(bbox=bbox, size=size, tile_patterns=tile_patterns,
                        grid_size=grid_size
                        )

    def minimal_meta_tile(self, tiles):
        """
        Returns a MetaTile that contains all `tiles` plus ``meta_buffer``,
        but nothing more.
        """

        tiles, grid_size, bounds = self._full_tile_list(tiles)
        tiles = list(tiles)
        bbox, buffers = self._meta_bbox(tiles=bounds)

        level = tiles[0][2]
        size = self._size_from_buffered_bbox(bbox, level)

        tile_pattern = self._tiles_pattern(tiles=tiles, grid_size=grid_size, buffers=buffers)

        return MetaTile(
            bbox=bbox,
            size=size,
            tile_patterns=tile_pattern,
            grid_size=grid_size,
        )

    def _size_from_buffered_bbox(self, bbox, level):
        # meta_size * tile_size + 2*buffer does not work,
        # since the buffer can get truncated at the grid border
        res = self.grid.resolution(level)
        width = int(round((bbox[2] - bbox[0]) / res))
        height = int(round((bbox[3] - bbox[1]) / res))
        return width, height

    def _full_tile_list(self, tiles):
        """
        Return a complete list of all tiles that a minimal meta tile with `tiles` contains.

        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> mgrid._full_tile_list([(0, 0, 2), (1, 1, 2)])
        ([(0, 1, 2), (1, 1, 2), (0, 0, 2), (1, 0, 2)], (2, 2), ((0, 0, 2), (1, 1, 2)))
        """
        tile = tiles.pop()
        z = tile[2]
        minx = maxx = tile[0]
        miny = maxy = tile[1]

        for tile in tiles:
            x, y = tile[:2]
            minx = min(minx, x)
            maxx = max(maxx, x)
            miny = min(miny, y)
            maxy = max(maxy, y)

        grid_size = 1+maxx-minx, 1+maxy-miny

        if self.grid.flipped_y_axis:
            ys = range(miny, maxy+1)
        else:
            ys = range(maxy, miny-1, -1)
        xs = range(minx, maxx+1)

        bounds = (minx, miny, z), (maxx, maxy, z)

        return list(_create_tile_list(xs, ys, z, (maxx+1, maxy+1))), grid_size, bounds

    def main_tile(self, tile_coord):
        x, y, z = tile_coord

        meta_size = self._meta_size(z)

        x0 = x//meta_size[0] * meta_size[0]
        y0 = y//meta_size[1] * meta_size[1]

        return x0, y0, z

    def tile_list(self, main_tile):
        tile_grid = self._meta_size(main_tile[2])
        return self._meta_tile_list(main_tile, tile_grid)

    def _meta_tile_list(self, main_tile, tile_grid):
        """
        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> mgrid._meta_tile_list((0, 1, 3), (2, 2))
        [(0, 1, 3), (1, 1, 3), (0, 0, 3), (1, 0, 3)]
        """
        minx, miny, z = self.main_tile(main_tile)
        maxx = minx + tile_grid[0] - 1
        maxy = miny + tile_grid[1] - 1
        if self.grid.flipped_y_axis:
            ys = range(miny, maxy+1)
        else:
            ys = range(maxy, miny-1, -1)
        xs = range(minx, maxx+1)

        return list(_create_tile_list(xs, ys, z, self.grid.grid_sizes[z]))

    def _tiles_pattern(self, grid_size, buffers, tile=None, tiles=None):
        """
        Returns the tile pattern for the given list of tiles.
        The result contains for each tile the ``tile_coord`` and the upper-left
        pixel coordinate of the tile in the meta tile image.

        >>> mgrid = MetaGrid(grid=TileGrid(), meta_size=(2, 2))
        >>> tiles = list(mgrid._tiles_pattern(tiles=[(0, 1, 2), (1, 1, 2)],
        ...                                   grid_size=(2, 1),
        ...                                   buffers=(0, 0, 10, 10)))
        >>> tiles[0], tiles[-1]
        (((0, 1, 2), (0, 10)), ((1, 1, 2), (256, 10)))

        >>> tiles = list(mgrid._tiles_pattern(tile=(1, 1, 2),
        ...                                   grid_size=(2, 2),
        ...                                   buffers=(10, 20, 30, 40)))
        >>> tiles[0], tiles[-1]
        (((0, 1, 2), (10, 40)), ((1, 0, 2), (266, 296)))

        """
        if tile:
            tiles = self._meta_tile_list(tile, grid_size)

        for i in range(grid_size[1]):
            for j in range(grid_size[0]):
                yield tiles[j+i*grid_size[0]], (
                    j*self.grid.tile_size[0] + buffers[0],
                    i*self.grid.tile_size[1] + buffers[3])

    def _meta_size(self, level):
        grid_size = self.grid.grid_sizes[level]
        return min(self.meta_size[0], grid_size[0]), min(self.meta_size[1], grid_size[1])

    def get_affected_level_tiles(self, bbox, level):
        """
        Get a list with all affected tiles for a `bbox` in the given `level`.

        :returns: the bbox, the size and a list with tile coordinates, sorted row-wise
        :rtype: ``bbox, (xs, yz), [(x, y, z), ...]``

        >>> grid = MetaGrid(TileGrid(), (2, 2))
        >>> bbox = (-20037508.34, -20037508.34, 20037508.34, 20037508.34)
        >>> grid.get_affected_level_tiles(bbox, 0)
        ... #doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
        ((-20037508.342789244, -20037508.342789244,\
          20037508.342789244, 20037508.342789244), (1, 1),\
          <generator object ...>)
        """

        # remove 1/10 of a pixel so we don't get a tiles we only touch
        delta = self.grid.resolutions[level] / 10.0
        x0, y0, _ = self.grid.tile(bbox[0]+delta, bbox[1]+delta, level)
        x1, y1, _ = self.grid.tile(bbox[2]-delta, bbox[3]-delta, level)

        meta_size = self._meta_size(level)

        x0 = x0//meta_size[0] * meta_size[0]
        x1 = x1//meta_size[0] * meta_size[0]
        y0 = y0//meta_size[1] * meta_size[1]
        y1 = y1//meta_size[1] * meta_size[1]

        try:
            return self._tile_iter(x0, y0, x1, y1, level)
        except IndexError:
            raise GridError('Invalid BBOX')

    def _tile_iter(self, x0, y0, x1, y1, level):
        meta_size = self._meta_size(level)

        xs = list(range(x0, x1+1, meta_size[0]))
        if self.grid.flipped_y_axis:
            y0, y1 = y1, y0
            ys = list(range(y0, y1+1, meta_size[1]))
        else:
            ys = list(range(y1, y0-1, -meta_size[1]))

        ll = (xs[0], ys[-1], level)
        ur = (xs[-1], ys[0], level)
        # add meta_size to get full affected bbox
        ur = ur[0]+meta_size[0]-1, ur[1]+meta_size[1]-1, ur[2]
        abbox = self.grid._tiles_bbox([ll, ur])
        return (abbox, (len(xs), len(ys)),
                _create_tile_list(xs, ys, level, self.grid.grid_sizes[level]))


class MetaTile(object):
    def __init__(self, bbox, size, tile_patterns, grid_size):
        self.bbox = bbox
        self.size = size
        self.tile_patterns = list(tile_patterns)
        self.grid_size = grid_size

    @property
    def tiles(self):
        return [t[0] for t in self.tile_patterns]

    @property
    def main_tile_coord(self):
        """
        Returns the "main" tile of the meta tile. This tile(coord) can be used
        for locking.

        >>> t = MetaTile(None, None, [((0, 0, 0), (0, 0)), ((1, 0, 0), (100, 0))], (2, 1))
        >>> t.main_tile_coord
        (0, 0, 0)
        >>> t = MetaTile(None, None, [(None, None), ((1, 0, 0), (100, 0))], (2, 1))
        >>> t.main_tile_coord
        (1, 0, 0)
        """
        for t in self.tiles:
            if t is not None:
                return t

    def __repr__(self):
        return "MetaTile(%r, %r, %r, %r)" % (self.bbox, self.size, self.grid_size,
                                             self.tile_patterns)


def bbox_intersects(one, two):
    a_x0, a_y0, a_x1, a_y1 = one
    b_x0, b_y0, b_x1, b_y1 = two

    if (
        a_x0 < b_x1 and
        a_x1 > b_x0 and
        a_y0 < b_y1 and
        a_y1 > b_y0
    ):
        return True

    return False


def bbox_contains(one, two):
    """
    Returns ``True`` if `one` contains `two`.

    >>> bbox_contains([0, 0, 10, 10], [2, 2, 4, 4])
    True
    >>> bbox_contains([0, 0, 10, 10], [0, 0, 11, 10])
    False

    Allow tiny rounding errors:

    >>> bbox_contains([0, 0, 10, 10], [0.000001, 0.0000001, 10.000001, 10.000001])
    False
    >>> bbox_contains([0, 0, 10, 10], [0.0000000000001, 0.0000000000001, 10.0000000000001, 10.0000000000001])
    True
    """
    a_x0, a_y0, a_x1, a_y1 = one
    b_x0, b_y0, b_x1, b_y1 = two

    x_delta = abs(a_x1 - a_x0) / 10e12
    y_delta = abs(a_y1 - a_y0) / 10e12

    if (
        a_x0 <= b_x0 + x_delta and
        a_x1 >= b_x1 - x_delta and
        a_y0 <= b_y0 + y_delta and
        a_y1 >= b_y1 - y_delta
    ):
        return True

    return False


def deg_to_m(deg):
    return deg * (6378137 * 2 * math.pi) / 360


OGC_PIXEL_SIZE = 0.00028  # m/px


def ogc_scale_to_res(scale):
    return scale * OGC_PIXEL_SIZE


def res_to_ogc_scale(res):
    return res / OGC_PIXEL_SIZE


def resolution_range(min_res=None, max_res=None, max_scale=None, min_scale=None):
    if min_scale == max_scale == min_res == max_res is None:
        return None
    if min_res or max_res:
        if not max_scale and not min_scale:
            return ResolutionRange(min_res, max_res)
    elif max_scale or min_scale:
        if not min_res and not max_res:
            min_res = ogc_scale_to_res(max_scale)
            max_res = ogc_scale_to_res(min_scale)
            return ResolutionRange(min_res, max_res)

    raise ValueError('requires either min_res/max_res or max_scale/min_scale')


class ResolutionRange(object):
    def __init__(self, min_res, max_res):
        self.min_res = min_res
        self.max_res = max_res

        if min_res and max_res:
            assert min_res > max_res

    def scale_denominator(self):
        min_scale = res_to_ogc_scale(self.max_res) if self.max_res else None
        max_scale = res_to_ogc_scale(self.min_res) if self.min_res else None
        return min_scale, max_scale

    def scale_hint(self):
        """
        Returns the min and max diagonal resolution.
        """
        min_res = self.min_res
        max_res = self.max_res
        if min_res:
            min_res = math.sqrt(2*min_res**2)
        if max_res:
            max_res = math.sqrt(2*max_res**2)
        return min_res, max_res

    def contains(self, bbox, size, srs):
        width, height = bbox_size(bbox)
        if srs.is_latlong:
            width = deg_to_m(width)
            height = deg_to_m(height)

        x_res = width/size[0]
        y_res = height/size[1]

        if self.min_res:
            min_res = self.min_res + 1e-6
            if min_res <= x_res or min_res <= y_res:
                return False
        if self.max_res:
            max_res = self.max_res
            if max_res > x_res or max_res > y_res:
                return False

        return True

    def __eq__(self, other):
        if not isinstance(other, ResolutionRange):
            return NotImplemented

        return (self.min_res == other.min_res
                and self.max_res == other.max_res)

    def __ne__(self, other):
        if not isinstance(other, ResolutionRange):
            return NotImplemented
        return not self == other

    def __repr__(self):
        return '<ResolutionRange(min_res=%.3f, max_res=%.3f)>' % (
            self.min_res or 9e99, self.max_res or 0)


def max_with_none(a, b):
    if a is None or b is None:
        return None
    else:
        return max(a, b)


def min_with_none(a, b):
    if a is None or b is None:
        return None
    else:
        return min(a, b)


def merge_resolution_range(a, b):
    if a and b:
        return resolution_range(min_res=max_with_none(a.min_res, b.min_res),
                                max_res=min_with_none(a.max_res, b.max_res))
    return None
